﻿using System;
using System.Collections.Generic;
using UnityEngine;
using Random = System.Random;

namespace Dub {

	public class Force : MVRScript {
        private Rigidbody _rb;
        private Random random;
        private void SyncReceiver(string receiver) {
			if (receiver != null)
            {
                ForceReceiver fr;
                _rb = ReceiverNameToForceReceiver.TryGetValue(receiver, out fr) ? fr.GetComponent<Rigidbody>() : null;
            } else {
				_rb = null;
			}
		}

        protected JSONStorableStringChooser ReceiverChoiceJSON;

        private ForceProducerV2.AxisName _forceAxis = ForceProducerV2.AxisName.Z;

        protected void SyncForceAxis(string axisName) {
			try {
				ForceProducerV2.AxisName an = (ForceProducerV2.AxisName)System.Enum.Parse(typeof(ForceProducerV2.AxisName), axisName);
				_forceAxis = an;
			}
			catch (System.ArgumentException) {
				Debug.LogError("Attempt to set axis to " + axisName + " which is not a valid axis name");
			}
		}

        protected JSONStorableStringChooser ForceAxisJSON;

		private ForceProducerV2.AxisName _torqueAxis = ForceProducerV2.AxisName.X;
		protected void SyncTorqueAxis(string axisName) {
			try {
				ForceProducerV2.AxisName an = (ForceProducerV2.AxisName)System.Enum.Parse(typeof(ForceProducerV2.AxisName), axisName);
				_torqueAxis = an;
			}
			catch (System.ArgumentException) {
				Debug.LogError("Attempt to set axis to " + axisName + " which is not a valid axis name");
			}
		}
		protected JSONStorableStringChooser TorqueAxisJSON;

        protected JSONStorableFloat ManualForceJSON;
        protected JSONStorableFloat ManualTorqueJSON;
        protected JSONStorableFloat PeriodJSON;
		protected JSONStorableFloat PeriodRatioJSON;
		protected JSONStorableFloat ForceDurationJSON;
		protected JSONStorableFloat ForceFactorJSON;
        protected JSONStorableFloat ForceFactorRandomVarianceJSON;
		protected JSONStorableFloat ForceQuicknessJSON;
        protected JSONStorableFloat ForceQuicknessRandomVarianceJSON;
        protected JSONStorableFloat TorqueFactorJSON;
        protected JSONStorableFloat TorqueFactorRandomVarianceJSON;
        protected JSONStorableFloat TorqueQuicknessJSON;
        protected JSONStorableFloat TorqueQuicknessRandomVarianceJSON;
        protected JSONStorableBool AutoForceJSON;
        protected JSONStorableBool AutoTorqueJSON;
        protected JSONStorableBool ApplyForceOnReturnJSON;

		protected List<string> ReceiverChoices;
		protected Dictionary<string, ForceReceiver> ReceiverNameToForceReceiver;

		public override void Init() {
			try
            {
                random = new Random();
				ReceiverChoices = new List<string>();
				ReceiverNameToForceReceiver = new Dictionary<string, ForceReceiver>();
				foreach (var fr in containingAtom.forceReceivers) {
					ReceiverChoices.Add(fr.name);
					ReceiverNameToForceReceiver.Add(fr.name, fr);
				}
				ReceiverChoiceJSON = new JSONStorableStringChooser("receiver", ReceiverChoices, null, "Receiver", SyncReceiver)
                    {
                        storeType = JSONStorableParam.StoreType.Full
                    };
                RegisterStringChooser(ReceiverChoiceJSON);
				var dp = CreateFilterablePopup(ReceiverChoiceJSON);
				dp.popupPanelHeight = 1100f;
				dp.popup.alwaysOpen = true;

				var axisChoices = Enum.GetNames(typeof(ForceProducerV2.AxisName));
				var axisChoicesList = new List<string>(axisChoices);

				ForceAxisJSON = new JSONStorableStringChooser("forceAxis", axisChoicesList, _forceAxis.ToString(), "Force Axis", SyncForceAxis)
                    {
                        storeType = JSONStorableParam.StoreType.Full
                    };
                RegisterStringChooser(ForceAxisJSON);
				CreatePopup(ForceAxisJSON,true);

				TorqueAxisJSON = new JSONStorableStringChooser("torqueAxis", axisChoicesList, _torqueAxis.ToString(), "Torque Axis", SyncTorqueAxis)
                    {
                        storeType = JSONStorableParam.StoreType.Full
                    };
                RegisterStringChooser(TorqueAxisJSON);
				CreatePopup(TorqueAxisJSON,true);

                PeriodJSON = new JSONStorableFloat("period", 0.5f, 0f, 10f, false)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(PeriodJSON);
				CreateSlider(PeriodJSON, true).label = "Period";

				PeriodRatioJSON = new JSONStorableFloat("periodRatio", 0.5f, 0f, 1f, true)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(PeriodRatioJSON);
				CreateSlider(PeriodRatioJSON, true).label = "Period Ratio";


				ForceDurationJSON = new JSONStorableFloat("forceDuration", 1f, 0f, 1f)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(ForceDurationJSON);
				CreateSlider(ForceDurationJSON, true).label = "Force Duration";;

                CreateSpacer(true).height = 15f;

				ForceFactorJSON = new JSONStorableFloat("forceFactor", 0f, 0f, 1000f, false)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(ForceFactorJSON);
				CreateSlider(ForceFactorJSON, true).label = "Force Factor";

                ForceFactorRandomVarianceJSON = new JSONStorableFloat("forceFactorRandomVariance", 0f, 0f, 500f, false)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(ForceFactorRandomVarianceJSON);
                CreateSlider(ForceFactorRandomVarianceJSON, true).label = "/\\ Random Variance";

                CreateSpacer(true).height = 15f;

				ForceQuicknessJSON = new JSONStorableFloat("forceQuickness", 10f, 0f, 50f, false, true)
                 {
                     storeType = JSONStorableParam.StoreType.Full
                 };
                RegisterFloat(ForceQuicknessJSON);
				CreateSlider(ForceQuicknessJSON, true).label = "Force Quickness";

                ForceQuicknessRandomVarianceJSON = new JSONStorableFloat("forceQuicknessRandomVariance", 0f, 0f, 25f, false)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(ForceQuicknessRandomVarianceJSON);
                CreateSlider(ForceQuicknessRandomVarianceJSON, true).label = "/\\ Random Variance";

                CreateSpacer(true).height = 15f;

				TorqueFactorJSON = new JSONStorableFloat("torqueFactor", 0f, 0f, 100f, false, true)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(TorqueFactorJSON);
				CreateSlider(TorqueFactorJSON, true).label = "Torque Factor";

                TorqueFactorRandomVarianceJSON = new JSONStorableFloat("torqueFactorRandomVariance", 0f, 0f, 50f, false)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(TorqueFactorRandomVarianceJSON);
                CreateSlider(TorqueFactorRandomVarianceJSON, true).label = "/\\ Random Variance";

                CreateSpacer(true).height = 15f;

				TorqueQuicknessJSON = new JSONStorableFloat("torqueQuickness", 10f, 0f, 50f, false, true)
                 {
                     storeType = JSONStorableParam.StoreType.Full
                 };
                RegisterFloat(TorqueQuicknessJSON);
				CreateSlider(TorqueQuicknessJSON, true).label = "Torque Quickness";

                TorqueQuicknessRandomVarianceJSON = new JSONStorableFloat("torqueQuicknessRandomVariance", 0f, 0f, 25f, false)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(TorqueQuicknessRandomVarianceJSON);
                CreateSlider(TorqueQuicknessRandomVarianceJSON, true).label = "/\\ Random Variance";

                CreateSpacer(true).height = 15f;

                ApplyForceOnReturnJSON = new JSONStorableBool("applyForceOnReturn", true)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterBool(ApplyForceOnReturnJSON);
				CreateToggle(ApplyForceOnReturnJSON, true).label = "Apply Force On Return?";;

                AutoForceJSON = new JSONStorableBool("autoForce", true)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterBool(AutoForceJSON);
                CreateToggle(AutoForceJSON, true).label = "Auto Force";

                AutoTorqueJSON = new JSONStorableBool("autoTorque", true)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterBool(AutoTorqueJSON);
                CreateToggle(AutoTorqueJSON, true).label = "Auto Torque";

                ManualForceJSON = new JSONStorableFloat("manualForce", 0f, 0f, 1f, true)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(ManualForceJSON);
                CreateSlider(ManualForceJSON, true).label = "Manual Force";;

                ManualTorqueJSON = new JSONStorableFloat("manualTorque", 0f, 0f, 1f, true)
                {
                    storeType = JSONStorableParam.StoreType.Full
                };
                RegisterFloat(ManualTorqueJSON);
                CreateSlider(ManualTorqueJSON, true).label = "Manual Torque";
			}
			catch (Exception e) {
				SuperController.LogError("Exception caught: " + e);
			}
		}

		protected float timer;
		protected float forceTimer;
		protected float flip;
		protected Vector3 targetForce;
		protected Vector3 targetTorque;
		protected Vector3 currentForce;
		protected Vector3 currentTorque;

		protected void Start() {
			timer = 0f;
			forceTimer = 0f;
			flip = 1f;
            if (ReceiverChoiceJSON.val == null) ReceiverChoiceJSON.val = "hip";
        }

		protected virtual Vector3 AxisToVector(ForceProducerV2.AxisName axis) {
			Vector3 result;
			if (_rb != null) {
				switch (axis) {
					case ForceProducerV2.AxisName.X:
						result = _rb.transform.right;
						break;
					case ForceProducerV2.AxisName.NegX:
						result = -_rb.transform.right;
						break;
					case ForceProducerV2.AxisName.Y:
						result = _rb.transform.up;
						break;
					case ForceProducerV2.AxisName.NegY:
						result = -_rb.transform.up;
						break;
					case ForceProducerV2.AxisName.Z:
						result = _rb.transform.forward;
						break;
					case ForceProducerV2.AxisName.NegZ:
						result = -_rb.transform.forward;
						break;
					default:
						result = Vector3.zero;
						break;
				}
			} else {
				result = Vector3.zero;
			}
			return (result);
		}

		protected void SetTargets(float percent)
        {
            var randomForceVariance = random.Next(
                    (int) -ForceFactorRandomVarianceJSON.val,
                    (int) ForceFactorRandomVarianceJSON.val
                );
            var randomTorqueVariance = random.Next(
                (int) -TorqueFactorRandomVarianceJSON.val,
                (int) TorqueFactorRandomVarianceJSON.val
            );
			if (AutoForceJSON.val) targetForce = AxisToVector(_forceAxis) * (percent * (ForceFactorJSON.val + randomForceVariance));
            if (AutoTorqueJSON.val) targetTorque = AxisToVector(_torqueAxis) * (percent * (TorqueFactorJSON.val + randomTorqueVariance));
		}

		// Use Update for the timers since this can account for time scale
		protected void Update()
        {
            if (!AutoForceJSON.val && !AutoTorqueJSON.val) return;
            timer -= Time.deltaTime;
            forceTimer -= Time.deltaTime;

            if (timer < 0.0f) {
				if ((flip > 0f && PeriodRatioJSON.val != 1f) || PeriodRatioJSON.val == 0f) {
					if (ApplyForceOnReturnJSON.val) {
						flip = -1f;
					} else {
						flip = 0f;
					}
                    timer = PeriodJSON.val * (1f - PeriodRatioJSON.val);
                    forceTimer = ForceDurationJSON.val * PeriodJSON.val;
				} else {
					flip = 1f;
                    timer = PeriodJSON.val * PeriodRatioJSON.val;
                    forceTimer = ForceDurationJSON.val * PeriodJSON.val;
				}
				SetTargets(flip);
			} else if (forceTimer < 0.0f) {
				SetTargets(0f);
			}
		}

		// FixedUpdate is called with each physics simulation frame by Unity
		void FixedUpdate() {
			try {
				// apply forces here
                var timeFactor = Time.fixedDeltaTime;
                var randomForceQuicknessVariance = random.Next(
                    (int) -ForceQuicknessRandomVarianceJSON.val,
                    (int) ForceQuicknessRandomVarianceJSON.val
                );
                var randomTorqueQuicknessVariance = random.Next(
                    (int) -TorqueQuicknessRandomVarianceJSON.val,
                    (int) TorqueQuicknessRandomVarianceJSON.val
                );
                currentForce = AutoForceJSON.val
                    ? Vector3.Lerp(currentForce, targetForce, timeFactor * (ForceQuicknessJSON.val + randomForceQuicknessVariance))
                    : AxisToVector(_forceAxis) * (ForceFactorJSON.val * ManualForceJSON.val * (ForceQuicknessJSON.val / 10));
                currentTorque = AutoTorqueJSON.val
                    ? Vector3.Lerp(currentTorque, targetTorque, timeFactor * (TorqueQuicknessJSON.val + randomTorqueQuicknessVariance))
                    : AxisToVector(_torqueAxis) * (TorqueFactorJSON.val * ManualTorqueJSON.val * (TorqueQuicknessJSON.val / 10));;

                if (_rb && (!SuperController.singleton || !SuperController.singleton.freezeAnimation)) {
					_rb.AddForce(currentForce, ForceMode.Force);
					_rb.AddTorque(currentTorque, ForceMode.Force);
				}
			}
			catch (Exception e) {
				SuperController.LogError("Exception caught: " + e);
			}
		}

	}
}

